@@ -6,7 +6,7 @@ import (
 	"testing"
 	"time"
 
-	"github.com/hashicorp/consul/testutil/retry"
+	"github.com/hashicorp/consul/sdk/testutil/retry"
 
 	"github.com/hashicorp/consul/agent/cache"
 	"github.com/hashicorp/consul/agent/connect"
@@ -1016,3 +1016,10 @@ func (r *testGatedRootsRPC) RPC(method string, args interface{}, reply interface
 	*replyReal = <-r.ValueCh
 	return nil
 }
+
+func TestConnectCALeaf_Key(t *testing.T) {
+	r := ConnectCALeafRequest{Service: "web"}
+	require.Equal(t, "service:web", r.Key())
+	r = ConnectCALeafRequest{Agent: "abc"}
+	require.Equal(t, "agent:abc", r.Key())
+}
