@@ -17,7 +17,7 @@ import (
 
 	"github.com/hashicorp/consul/agent/mock"
 	"github.com/hashicorp/consul/api"
-	"github.com/hashicorp/consul/testutil/retry"
+	"github.com/hashicorp/consul/sdk/testutil/retry"
 	"github.com/hashicorp/consul/types"
 	uuid "github.com/hashicorp/go-uuid"
 )
@@ -44,11 +44,12 @@ func TestCheckMonitor_Script(t *testing.T) {
 		t.Run(tt.status, func(t *testing.T) {
 			notif := mock.NewNotify()
 			check := &CheckMonitor{
-				Notify:   notif,
-				CheckID:  types.CheckID("foo"),
-				Script:   tt.script,
-				Interval: 25 * time.Millisecond,
-				Logger:   log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+				Notify:        notif,
+				CheckID:       types.CheckID("foo"),
+				Script:        tt.script,
+				Interval:      25 * time.Millisecond,
+				Logger:        log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+				OutputMaxSize: DefaultBufSize,
 			}
 			check.Start()
 			defer check.Stop()
@@ -79,11 +80,12 @@ func TestCheckMonitor_Args(t *testing.T) {
 		t.Run(tt.status, func(t *testing.T) {
 			notif := mock.NewNotify()
 			check := &CheckMonitor{
-				Notify:     notif,
-				CheckID:    types.CheckID("foo"),
-				ScriptArgs: tt.args,
-				Interval:   25 * time.Millisecond,
-				Logger:     log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+				Notify:        notif,
+				CheckID:       types.CheckID("foo"),
+				ScriptArgs:    tt.args,
+				Interval:      25 * time.Millisecond,
+				Logger:        log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+				OutputMaxSize: DefaultBufSize,
 			}
 			check.Start()
 			defer check.Stop()
@@ -103,12 +105,13 @@ func TestCheckMonitor_Timeout(t *testing.T) {
 	// t.Parallel() // timing test. no parallel
 	notif := mock.NewNotify()
 	check := &CheckMonitor{
-		Notify:     notif,
-		CheckID:    types.CheckID("foo"),
-		ScriptArgs: []string{"sh", "-c", "sleep 1 && exit 0"},
-		Interval:   50 * time.Millisecond,
-		Timeout:    25 * time.Millisecond,
-		Logger:     log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+		Notify:        notif,
+		CheckID:       types.CheckID("foo"),
+		ScriptArgs:    []string{"sh", "-c", "sleep 1 && exit 0"},
+		Interval:      50 * time.Millisecond,
+		Timeout:       25 * time.Millisecond,
+		Logger:        log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+		OutputMaxSize: DefaultBufSize,
 	}
 	check.Start()
 	defer check.Stop()
@@ -128,11 +131,12 @@ func TestCheckMonitor_RandomStagger(t *testing.T) {
 	// t.Parallel() // timing test. no parallel
 	notif := mock.NewNotify()
 	check := &CheckMonitor{
-		Notify:     notif,
-		CheckID:    types.CheckID("foo"),
-		ScriptArgs: []string{"sh", "-c", "exit 0"},
-		Interval:   25 * time.Millisecond,
-		Logger:     log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+		Notify:        notif,
+		CheckID:       types.CheckID("foo"),
+		ScriptArgs:    []string{"sh", "-c", "exit 0"},
+		Interval:      25 * time.Millisecond,
+		Logger:        log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+		OutputMaxSize: DefaultBufSize,
 	}
 	check.Start()
 	defer check.Stop()
@@ -153,11 +157,12 @@ func TestCheckMonitor_LimitOutput(t *testing.T) {
 	t.Parallel()
 	notif := mock.NewNotify()
 	check := &CheckMonitor{
-		Notify:     notif,
-		CheckID:    types.CheckID("foo"),
-		ScriptArgs: []string{"od", "-N", "81920", "/dev/urandom"},
-		Interval:   25 * time.Millisecond,
-		Logger:     log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+		Notify:        notif,
+		CheckID:       types.CheckID("foo"),
+		ScriptArgs:    []string{"od", "-N", "81920", "/dev/urandom"},
+		Interval:      25 * time.Millisecond,
+		Logger:        log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+		OutputMaxSize: DefaultBufSize,
 	}
 	check.Start()
 	defer check.Stop()
@@ -165,7 +170,7 @@ func TestCheckMonitor_LimitOutput(t *testing.T) {
 	time.Sleep(50 * time.Millisecond)
 
 	// Allow for extra bytes for the truncation message
-	if len(notif.Output("foo")) > BufSize+100 {
+	if len(notif.Output("foo")) > DefaultBufSize+100 {
 		t.Fatalf("output size is too long")
 	}
 }
@@ -287,7 +292,7 @@ func TestCheckHTTP(t *testing.T) {
 				}
 
 				// Body larger than 4k limit
-				body := bytes.Repeat([]byte{'a'}, 2*BufSize)
+				body := bytes.Repeat([]byte{'a'}, 2*DefaultBufSize)
 				w.WriteHeader(tt.code)
 				w.Write(body)
 			}))
@@ -295,13 +300,14 @@ func TestCheckHTTP(t *testing.T) {
 
 			notif := mock.NewNotify()
 			check := &CheckHTTP{
-				Notify:   notif,
-				CheckID:  types.CheckID("foo"),
-				HTTP:     server.URL,
-				Method:   tt.method,
-				Header:   tt.header,
-				Interval: 10 * time.Millisecond,
-				Logger:   log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+				Notify:        notif,
+				CheckID:       types.CheckID("foo"),
+				HTTP:          server.URL,
+				Method:        tt.method,
+				OutputMaxSize: DefaultBufSize,
+				Header:        tt.header,
+				Interval:      10 * time.Millisecond,
+				Logger:        log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
 			}
 			check.Start()
 			defer check.Stop()
@@ -313,15 +319,52 @@ func TestCheckHTTP(t *testing.T) {
 				if got, want := notif.State("foo"), tt.status; got != want {
 					r.Fatalf("got state %q want %q", got, want)
 				}
-				// Allow slightly more data than BufSize, for the header
-				if n := len(notif.Output("foo")); n > (BufSize + 256) {
-					r.Fatalf("output too long: %d (%d-byte limit)", n, BufSize)
+				// Allow slightly more data than DefaultBufSize, for the header
+				if n := len(notif.Output("foo")); n > (DefaultBufSize + 256) {
+					r.Fatalf("output too long: %d (%d-byte limit)", n, DefaultBufSize)
 				}
 			})
 		})
 	}
 }
 
+func TestCheckMaxOutputSize(t *testing.T) {
+	t.Parallel()
+	timeout := 5 * time.Millisecond
+	server := httptest.NewServer(http.HandlerFunc(func(writer http.ResponseWriter, req *http.Request) {
+		body := bytes.Repeat([]byte{'x'}, 2*DefaultBufSize)
+		writer.WriteHeader(200)
+		writer.Write(body)
+	}))
+	defer server.Close()
+
+	notif := mock.NewNotify()
+	maxOutputSize := 32
+	check := &CheckHTTP{
+		Notify:        notif,
+		CheckID:       types.CheckID("bar"),
+		HTTP:          server.URL + "/v1/agent/self",
+		Timeout:       timeout,
+		Interval:      2 * time.Millisecond,
+		Logger:        log.New(ioutil.Discard, uniqueID(), log.LstdFlags),
+		OutputMaxSize: maxOutputSize,
+	}
+
+	check.Start()
+	defer check.Stop()
+	retry.Run(t, func(r *retry.R) {
+		if got, want := notif.Updates("bar"), 2; got < want {
+			r.Fatalf("got %d updates want at least %d", got, want)
+		}
+		if got, want := notif.State("bar"), api.HealthPassing; got != want {
+			r.Fatalf("got state %q want %q", got, want)
+		}
+		if got, want := notif.Output("bar"), "HTTP GET "+server.URL+"/v1/agent/self: 200 OK Output: "+strings.Repeat("x", maxOutputSize); got != want {
+			r.Fatalf("got state %q want %q", got, want)
+		}
+	})
+}
+
 func TestCheckHTTPTimeout(t *testing.T) {
 	t.Parallel()
 	timeout := 5 * time.Millisecond
@@ -372,7 +415,7 @@ func TestCheckHTTP_disablesKeepAlives(t *testing.T) {
 func largeBodyHandler(code int) http.Handler {
 	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 		// Body larger than 4k limit
-		body := bytes.Repeat([]byte{'a'}, 2*BufSize)
+		body := bytes.Repeat([]byte{'a'}, 2*DefaultBufSize)
 		w.WriteHeader(code)
 		w.Write(body)
 	})
@@ -517,6 +560,9 @@ func TestCheckTCPPassing(t *testing.T) {
 	if os.Getenv("TRAVIS") == "true" {
 		t.Skip("IPV6 not supported on travis-ci")
 	}
+	if os.Getenv("CIRCLECI") == "true" {
+		t.Skip("IPV6 not supported on CircleCI")
+	}
 	tcpServer = mockTCPServer(`tcp6`)
 	expectTCPStatus(t, tcpServer.Addr().String(), api.HealthPassing)
 	tcpServer.Close()
