@@ -7,6 +7,7 @@ import (
 	"os"
 	"time"
 
+	"github.com/hashicorp/consul/agent/checks"
 	"github.com/hashicorp/consul/agent/consul/autopilot"
 	"github.com/hashicorp/consul/agent/structs"
 	"github.com/hashicorp/consul/lib"
@@ -26,7 +27,7 @@ const (
 	DefaultWANSerfPort = 8302
 
 	// DefaultRaftMultiplier is used as a baseline Raft configuration that
-	// will be reliable on a very basic server. See docs/guides/performance.html
+	// will be reliable on a very basic server. See docs/install/performance.html
 	// for information on how this value was obtained.
 	DefaultRaftMultiplier uint = 5
 
@@ -147,6 +148,9 @@ type Config struct {
 	// leader election.
 	ReconcileInterval time.Duration
 
+	// LogLevel is the level of the logs to write. Defaults to "INFO".
+	LogLevel string
+
 	// LogOutput is the location to write logs to. If this is not set,
 	// logs will go to stderr.
 	LogOutput io.Writer
@@ -243,6 +247,11 @@ type Config struct {
 	// a substantial cost.
 	ACLPolicyTTL time.Duration
 
+	// ACLRoleTTL controls the time-to-live of cached ACL roles.
+	// It can be set to zero to disable caching, but this adds
+	// a substantial cost.
+	ACLRoleTTL time.Duration
+
 	// ACLDisabledTTL is the time between checking if ACLs should be
 	// enabled. This
 	ACLDisabledTTL time.Duration
@@ -313,6 +322,16 @@ type Config struct {
 	// Minimum Session TTL
 	SessionTTLMin time.Duration
 
+	// maxTokenExpirationDuration is the maximum difference allowed between
+	// ACLToken CreateTime and ExpirationTime values if ExpirationTime is set
+	// on a token.
+	ACLTokenMaxExpirationTTL time.Duration
+
+	// ACLTokenMinExpirationTTL is the minimum difference allowed between
+	// ACLToken CreateTime and ExpirationTime values if ExpirationTime is set
+	// on a token.
+	ACLTokenMinExpirationTTL time.Duration
+
 	// ServerUp callback can be used to trigger a notification that
 	// a Consul server is now up and known about.
 	ServerUp func()
@@ -321,6 +340,20 @@ type Config struct {
 	// user events. This function should not block.
 	UserEventHandler func(serf.UserEvent)
 
+	// ConfigReplicationRate is the max number of replication rounds that can
+	// be run per second. Note that either 1 or 2 RPCs are used during each replication
+	// round
+	ConfigReplicationRate int
+
+	// ConfigReplicationBurst is how many replication rounds can be bursted after a
+	// period of idleness
+	ConfigReplicationBurst int
+
+	// ConfigReplicationApply limit is the max number of replication-related
+	// apply operations that we allow during a one second period. This is
+	// used to limit the amount of Raft bandwidth used for replication.
+	ConfigReplicationApplyLimit int
+
 	// CoordinateUpdatePeriod controls how long a server batches coordinate
 	// updates before applying them in a Raft transaction. A larger period
 	// leads to fewer Raft transactions, but also the stored coordinates
@@ -336,6 +369,9 @@ type Config struct {
 	// warning and discard the remaining updates.
 	CoordinateUpdateMaxBatches int
 
+	// CheckOutputMaxSize control the max size of output of checks
+	CheckOutputMaxSize int
+
 	// RPCHoldTimeout is how long an RPC can be "held" before it is errored.
 	// This is used to paper over a loss of leadership by instead holding RPCs,
 	// so that the caller experiences a slow response rather than an error.
@@ -377,17 +413,29 @@ type Config struct {
 	// CAConfig is used to apply the initial Connect CA configuration when
 	// bootstrapping.
 	CAConfig *structs.CAConfiguration
+
+	// ConfigEntryBootstrap contains a list of ConfigEntries to ensure are created
+	// If entries of the same Kind/Name exist already these will not update them.
+	ConfigEntryBootstrap []structs.ConfigEntry
+
+	// AutoEncryptAllowTLS is whether to enable the server responding to
+	// AutoEncrypt.Sign requests.
+	AutoEncryptAllowTLS bool
 }
 
+// ToTLSUtilConfig is only used by tests, usually the config is being passed
+// down from the agent.
 func (c *Config) ToTLSUtilConfig() tlsutil.Config {
 	return tlsutil.Config{
 		VerifyIncoming:           c.VerifyIncoming,
 		VerifyOutgoing:           c.VerifyOutgoing,
+		VerifyServerHostname:     c.VerifyServerHostname,
 		CAFile:                   c.CAFile,
 		CAPath:                   c.CAPath,
 		CertFile:                 c.CertFile,
 		KeyFile:                  c.KeyFile,
 		NodeName:                 c.NodeName,
+		Domain:                   c.Domain,
 		ServerName:               c.ServerName,
 		TLSMinVersion:            c.TLSMinVersion,
 		CipherSuites:             c.TLSCipherSuites,
@@ -432,26 +480,32 @@ func DefaultConfig() *Config {
 	}
 
 	conf := &Config{
-		Build:                    version.Version,
-		Datacenter:               DefaultDC,
-		NodeName:                 hostname,
-		RPCAddr:                  DefaultRPCAddr,
-		RaftConfig:               raft.DefaultConfig(),
-		SerfLANConfig:            lib.SerfDefaultConfig(),
-		SerfWANConfig:            lib.SerfDefaultConfig(),
-		SerfFloodInterval:        60 * time.Second,
-		ReconcileInterval:        60 * time.Second,
-		ProtocolVersion:          ProtocolVersion2Compatible,
-		ACLPolicyTTL:             30 * time.Second,
-		ACLTokenTTL:              30 * time.Second,
-		ACLDefaultPolicy:         "allow",
-		ACLDownPolicy:            "extend-cache",
-		ACLReplicationRate:       1,
-		ACLReplicationBurst:      5,
-		ACLReplicationApplyLimit: 100, // ops / sec
-		TombstoneTTL:             15 * time.Minute,
-		TombstoneTTLGranularity:  30 * time.Second,
-		SessionTTLMin:            10 * time.Second,
+		Build:                       version.Version,
+		Datacenter:                  DefaultDC,
+		NodeName:                    hostname,
+		RPCAddr:                     DefaultRPCAddr,
+		RaftConfig:                  raft.DefaultConfig(),
+		SerfLANConfig:               lib.SerfDefaultConfig(),
+		SerfWANConfig:               lib.SerfDefaultConfig(),
+		SerfFloodInterval:           60 * time.Second,
+		ReconcileInterval:           60 * time.Second,
+		ProtocolVersion:             ProtocolVersion2Compatible,
+		ACLRoleTTL:                  30 * time.Second,
+		ACLPolicyTTL:                30 * time.Second,
+		ACLTokenTTL:                 30 * time.Second,
+		ACLDefaultPolicy:            "allow",
+		ACLDownPolicy:               "extend-cache",
+		ACLReplicationRate:          1,
+		ACLReplicationBurst:         5,
+		ACLReplicationApplyLimit:    100, // ops / sec
+		ConfigReplicationRate:       1,
+		ConfigReplicationBurst:      5,
+		ConfigReplicationApplyLimit: 100, // ops / sec
+		TombstoneTTL:                15 * time.Minute,
+		TombstoneTTLGranularity:     30 * time.Second,
+		SessionTTLMin:               10 * time.Second,
+		ACLTokenMinExpirationTTL:    1 * time.Minute,
+		ACLTokenMaxExpirationTTL:    24 * time.Hour,
 
 		// These are tuned to provide a total throughput of 128 updates
 		// per second. If you update these, you should update the client-
@@ -460,6 +514,8 @@ func DefaultConfig() *Config {
 		CoordinateUpdateBatchSize:  128,
 		CoordinateUpdateMaxBatches: 5,
 
+		CheckOutputMaxSize: checks.DefaultBufSize,
+
 		RPCRate:     rate.Inf,
 		RPCMaxBurst: 1000,
 
@@ -498,6 +554,10 @@ func DefaultConfig() *Config {
 	conf.SerfLANConfig.MemberlistConfig.BindPort = DefaultLANSerfPort
 	conf.SerfWANConfig.MemberlistConfig.BindPort = DefaultWANSerfPort
 
+	// Allow dead nodes to be replaced after 30 seconds.
+	conf.SerfLANConfig.MemberlistConfig.DeadNodeReclaimTime = 30 * time.Second
+	conf.SerfWANConfig.MemberlistConfig.DeadNodeReclaimTime = 30 * time.Second
+
 	// Raft protocol version 3 only works with other Consul servers running
 	// 0.8.0 or later.
 	conf.RaftConfig.ProtocolVersion = 3
