@@ -21,7 +21,7 @@ type KVS struct {
 // preApply does all the verification of a KVS update that is performed BEFORE
 // we submit as a Raft log entry. This includes enforcing the lock delay which
 // must only be done on the leader.
-func kvsPreApply(srv *Server, rule acl.ACL, op api.KVOp, dirEnt *structs.DirEntry) (bool, error) {
+func kvsPreApply(srv *Server, rule acl.Authorizer, op api.KVOp, dirEnt *structs.DirEntry) (bool, error) {
 	// Verify the entry.
 
 	if dirEnt.Key == "" && op != api.KVDeleteTree {
@@ -81,11 +81,10 @@ func (k *KVS) Apply(args *structs.KVSRequest, reply *bool) error {
 	if done, err := k.srv.forward("KVS.Apply", args, args, reply); done {
 		return err
 	}
-	defer metrics.MeasureSince([]string{"consul", "kvs", "apply"}, time.Now())
 	defer metrics.MeasureSince([]string{"kvs", "apply"}, time.Now())
 
 	// Perform the pre-apply checks.
-	acl, err := k.srv.resolveToken(args.Token)
+	acl, err := k.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
@@ -121,7 +120,7 @@ func (k *KVS) Get(args *structs.KeyRequest, reply *structs.IndexedDirEntries) er
 		return err
 	}
 
-	aclRule, err := k.srv.resolveToken(args.Token)
+	aclRule, err := k.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
@@ -160,7 +159,7 @@ func (k *KVS) List(args *structs.KeyRequest, reply *structs.IndexedDirEntries) e
 		return err
 	}
 
-	aclToken, err := k.srv.resolveToken(args.Token)
+	aclToken, err := k.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
@@ -204,7 +203,7 @@ func (k *KVS) ListKeys(args *structs.KeyListRequest, reply *structs.IndexedKeyLi
 		return err
 	}
 
-	aclToken, err := k.srv.resolveToken(args.Token)
+	aclToken, err := k.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
