@@ -2,21 +2,25 @@ package consul
 
 import (
 	"os"
+	"reflect"
 	"testing"
 	"time"
 
+	"github.com/hashicorp/consul/agent/connect"
 	"github.com/hashicorp/consul/agent/structs"
 	"github.com/hashicorp/consul/api"
+	"github.com/hashicorp/consul/sdk/testutil/retry"
 	"github.com/hashicorp/consul/testrpc"
-	"github.com/hashicorp/consul/testutil/retry"
 	"github.com/hashicorp/net-rpc-msgpackrpc"
 	"github.com/hashicorp/serf/serf"
+	"github.com/stretchr/testify/require"
 )
 
 func TestLeader_RegisterMember(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "deny"
 		c.ACLEnforceVersion8 = true
@@ -86,6 +90,7 @@ func TestLeader_FailedMember(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "deny"
 		c.ACLEnforceVersion8 = true
@@ -147,6 +152,7 @@ func TestLeader_LeftMember(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "deny"
 		c.ACLEnforceVersion8 = true
@@ -193,6 +199,7 @@ func TestLeader_ReapMember(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "deny"
 		c.ACLEnforceVersion8 = true
@@ -254,6 +261,7 @@ func TestLeader_ReapServer(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "allow"
 		c.ACLEnforceVersion8 = true
@@ -264,6 +272,7 @@ func TestLeader_ReapServer(t *testing.T) {
 
 	dir2, s2 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "allow"
 		c.ACLEnforceVersion8 = true
@@ -274,6 +283,7 @@ func TestLeader_ReapServer(t *testing.T) {
 
 	dir3, s3 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "allow"
 		c.ACLEnforceVersion8 = true
@@ -287,6 +297,8 @@ func TestLeader_ReapServer(t *testing.T) {
 	joinLAN(t, s1, s3)
 
 	testrpc.WaitForLeader(t, s1.RPC, "dc1")
+	testrpc.WaitForLeader(t, s2.RPC, "dc1")
+	testrpc.WaitForLeader(t, s3.RPC, "dc1")
 	state := s1.fsm.State()
 
 	// s3 should be registered
@@ -327,6 +339,7 @@ func TestLeader_Reconcile_ReapMember(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "deny"
 		c.ACLEnforceVersion8 = true
@@ -376,6 +389,7 @@ func TestLeader_Reconcile(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "deny"
 		c.ACLEnforceVersion8 = true
@@ -705,6 +719,7 @@ func TestLeader_ReapTombstones(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
 		c.ACLDatacenter = "dc1"
+		c.ACLsEnabled = true
 		c.ACLMasterToken = "root"
 		c.ACLDefaultPolicy = "deny"
 		c.TombstoneTTL = 50 * time.Millisecond
@@ -860,7 +875,6 @@ func TestLeader_RollRaftServer(t *testing.T) {
 }
 
 func TestLeader_ChangeServerID(t *testing.T) {
-	t.Parallel()
 	conf := func(c *Config) {
 		c.Bootstrap = false
 		c.BootstrapExpect = 3
@@ -885,6 +899,7 @@ func TestLeader_ChangeServerID(t *testing.T) {
 	joinLAN(t, s2, s1)
 	joinLAN(t, s3, s1)
 	for _, s := range servers {
+		testrpc.WaitForTestAgent(t, s.RPC, "dc1")
 		retry.Run(t, func(r *retry.R) { r.Check(wantPeers(s, 3)) })
 	}
 
@@ -916,6 +931,8 @@ func TestLeader_ChangeServerID(t *testing.T) {
 	defer os.RemoveAll(dir4)
 	defer s4.Shutdown()
 	joinLAN(t, s4, s1)
+	testrpc.WaitForTestAgent(t, s1.RPC, "dc1")
+	testrpc.WaitForTestAgent(t, s4.RPC, "dc1")
 	servers[2] = s4
 
 	// While integrating #3327 it uncovered that this test was flaky. The
@@ -935,6 +952,69 @@ func TestLeader_ChangeServerID(t *testing.T) {
 	})
 }
 
+func TestLeader_ChangeNodeID(t *testing.T) {
+	t.Parallel()
+	dir1, s1 := testServer(t)
+	defer os.RemoveAll(dir1)
+	defer s1.Shutdown()
+
+	dir2, s2 := testServerDCBootstrap(t, "dc1", false)
+	defer os.RemoveAll(dir2)
+	defer s2.Shutdown()
+
+	dir3, s3 := testServerDCBootstrap(t, "dc1", false)
+	defer os.RemoveAll(dir3)
+	defer s3.Shutdown()
+
+	servers := []*Server{s1, s2, s3}
+
+	// Try to join and wait for all servers to get promoted
+	joinLAN(t, s2, s1)
+	joinLAN(t, s3, s1)
+	for _, s := range servers {
+		testrpc.WaitForTestAgent(t, s.RPC, "dc1")
+		retry.Run(t, func(r *retry.R) { r.Check(wantPeers(s, 3)) })
+	}
+
+	// Shut down a server, freeing up its address/port
+	s3.Shutdown()
+
+	retry.Run(t, func(r *retry.R) {
+		failed := 0
+		for _, m := range s1.LANMembers() {
+			if m.Status == serf.StatusFailed {
+				failed++
+			}
+		}
+		require.Equal(r, 1, failed)
+	})
+
+	// Bring up a new server with s3's name that will get a different ID
+	dir4, s4 := testServerWithConfig(t, func(c *Config) {
+		c.Bootstrap = false
+		c.Datacenter = "dc1"
+		c.NodeName = s3.config.NodeName
+	})
+	defer os.RemoveAll(dir4)
+	defer s4.Shutdown()
+	joinLAN(t, s4, s1)
+	servers[2] = s4
+
+	// Make sure the dead server is gone from both Raft and Serf and we're back to 3 total peers
+	retry.Run(t, func(r *retry.R) {
+		r.Check(wantRaft(servers))
+		for _, s := range servers {
+			r.Check(wantPeers(s, 3))
+		}
+	})
+
+	retry.Run(t, func(r *retry.R) {
+		for _, m := range s1.LANMembers() {
+			require.Equal(r, serf.StatusAlive, m.Status)
+		}
+	})
+}
+
 func TestLeader_ACL_Initialization(t *testing.T) {
 	t.Parallel()
 
@@ -942,13 +1022,12 @@ func TestLeader_ACL_Initialization(t *testing.T) {
 		name      string
 		build     string
 		master    string
-		init      bool
 		bootstrap bool
 	}{
-		{"old version, no master", "0.8.0", "", false, false},
-		{"old version, master", "0.8.0", "root", false, false},
-		{"new version, no master", "0.9.1", "", true, true},
-		{"new version, master", "0.9.1", "root", true, false},
+		{"old version, no master", "0.8.0", "", true},
+		{"old version, master", "0.8.0", "root", false},
+		{"new version, no master", "0.9.1", "", true},
+		{"new version, master", "0.9.1", "root", false},
 	}
 	for _, tt := range tests {
 		t.Run(tt.name, func(t *testing.T) {
@@ -957,47 +1036,312 @@ func TestLeader_ACL_Initialization(t *testing.T) {
 				c.Bootstrap = true
 				c.Datacenter = "dc1"
 				c.ACLDatacenter = "dc1"
+				c.ACLsEnabled = true
 				c.ACLMasterToken = tt.master
 			}
 			dir1, s1 := testServerWithConfig(t, conf)
 			defer os.RemoveAll(dir1)
 			defer s1.Shutdown()
-			testrpc.WaitForLeader(t, s1.RPC, "dc1")
+			testrpc.WaitForTestAgent(t, s1.RPC, "dc1")
 
 			if tt.master != "" {
-				_, master, err := s1.fsm.State().ACLGet(nil, tt.master)
-				if err != nil {
-					t.Fatalf("err: %v", err)
-				}
-				if master == nil {
-					t.Fatalf("master token wasn't created")
-				}
+				_, master, err := s1.fsm.State().ACLTokenGetBySecret(nil, tt.master)
+				require.NoError(t, err)
+				require.NotNil(t, master)
 			}
 
-			_, anon, err := s1.fsm.State().ACLGet(nil, anonymousToken)
-			if err != nil {
-				t.Fatalf("err: %v", err)
-			}
-			if anon == nil {
-				t.Fatalf("anonymous token wasn't created")
-			}
+			_, anon, err := s1.fsm.State().ACLTokenGetBySecret(nil, anonymousToken)
+			require.NoError(t, err)
+			require.NotNil(t, anon)
 
-			bs, err := s1.fsm.State().ACLGetBootstrap()
-			if err != nil {
-				t.Fatalf("err: %v", err)
-			}
-			if !tt.init {
-				if bs != nil {
-					t.Fatalf("bootstrap should not be initialized")
-				}
-			} else {
-				if bs == nil {
-					t.Fatalf("bootstrap should be initialized")
-				}
-				if got, want := bs.AllowBootstrap, tt.bootstrap; got != want {
-					t.Fatalf("got %v want %v", got, want)
-				}
-			}
+			canBootstrap, _, err := s1.fsm.State().CanBootstrapACLToken()
+			require.NoError(t, err)
+			require.Equal(t, tt.bootstrap, canBootstrap)
+
+			_, policy, err := s1.fsm.State().ACLPolicyGetByID(nil, structs.ACLPolicyGlobalManagementID)
+			require.NoError(t, err)
+			require.NotNil(t, policy)
 		})
 	}
 }
+
+func TestLeader_CARootPruning(t *testing.T) {
+	t.Parallel()
+
+	caRootPruneInterval = 200 * time.Millisecond
+
+	require := require.New(t)
+	dir1, s1 := testServer(t)
+	defer os.RemoveAll(dir1)
+	defer s1.Shutdown()
+	codec := rpcClient(t, s1)
+	defer codec.Close()
+
+	testrpc.WaitForTestAgent(t, s1.RPC, "dc1")
+
+	// Get the current root
+	rootReq := &structs.DCSpecificRequest{
+		Datacenter: "dc1",
+	}
+	var rootList structs.IndexedCARoots
+	require.Nil(msgpackrpc.CallWithCodec(codec, "ConnectCA.Roots", rootReq, &rootList))
+	require.Len(rootList.Roots, 1)
+	oldRoot := rootList.Roots[0]
+
+	// Update the provider config to use a new private key, which should
+	// cause a rotation.
+	_, newKey, err := connect.GeneratePrivateKey()
+	require.NoError(err)
+	newConfig := &structs.CAConfiguration{
+		Provider: "consul",
+		Config: map[string]interface{}{
+			"LeafCertTTL":    "500ms",
+			"PrivateKey":     newKey,
+			"RootCert":       "",
+			"RotationPeriod": "2160h",
+			"SkipValidate":   true,
+		},
+	}
+	{
+		args := &structs.CARequest{
+			Datacenter: "dc1",
+			Config:     newConfig,
+		}
+		var reply interface{}
+
+		require.NoError(msgpackrpc.CallWithCodec(codec, "ConnectCA.ConfigurationSet", args, &reply))
+	}
+
+	// Should have 2 roots now.
+	_, roots, err := s1.fsm.State().CARoots(nil)
+	require.NoError(err)
+	require.Len(roots, 2)
+
+	time.Sleep(2 * time.Second)
+
+	// Now the old root should be pruned.
+	_, roots, err = s1.fsm.State().CARoots(nil)
+	require.NoError(err)
+	require.Len(roots, 1)
+	require.True(roots[0].Active)
+	require.NotEqual(roots[0].ID, oldRoot.ID)
+}
+
+func TestLeader_PersistIntermediateCAs(t *testing.T) {
+	t.Parallel()
+
+	require := require.New(t)
+	dir1, s1 := testServer(t)
+	defer os.RemoveAll(dir1)
+	defer s1.Shutdown()
+	codec := rpcClient(t, s1)
+	defer codec.Close()
+
+	dir2, s2 := testServerDCBootstrap(t, "dc1", false)
+	defer os.RemoveAll(dir2)
+	defer s2.Shutdown()
+
+	dir3, s3 := testServerDCBootstrap(t, "dc1", false)
+	defer os.RemoveAll(dir3)
+	defer s3.Shutdown()
+
+	joinLAN(t, s2, s1)
+	joinLAN(t, s3, s1)
+
+	testrpc.WaitForLeader(t, s1.RPC, "dc1")
+
+	// Get the current root
+	rootReq := &structs.DCSpecificRequest{
+		Datacenter: "dc1",
+	}
+	var rootList structs.IndexedCARoots
+	require.Nil(msgpackrpc.CallWithCodec(codec, "ConnectCA.Roots", rootReq, &rootList))
+	require.Len(rootList.Roots, 1)
+
+	// Update the provider config to use a new private key, which should
+	// cause a rotation.
+	_, newKey, err := connect.GeneratePrivateKey()
+	require.NoError(err)
+	newConfig := &structs.CAConfiguration{
+		Provider: "consul",
+		Config: map[string]interface{}{
+			"PrivateKey":     newKey,
+			"RootCert":       "",
+			"RotationPeriod": 90 * 24 * time.Hour,
+		},
+	}
+	{
+		args := &structs.CARequest{
+			Datacenter: "dc1",
+			Config:     newConfig,
+		}
+		var reply interface{}
+
+		require.NoError(msgpackrpc.CallWithCodec(codec, "ConnectCA.ConfigurationSet", args, &reply))
+	}
+
+	// Get the active root before leader change.
+	_, root := s1.getCAProvider()
+	require.Len(root.IntermediateCerts, 1)
+
+	// Force a leader change and make sure the root CA values are preserved.
+	s1.Leave()
+	s1.Shutdown()
+
+	retry.Run(t, func(r *retry.R) {
+		var leader *Server
+		for _, s := range []*Server{s2, s3} {
+			if s.IsLeader() {
+				leader = s
+				break
+			}
+		}
+		if leader == nil {
+			r.Fatal("no leader")
+		}
+
+		_, newLeaderRoot := leader.getCAProvider()
+		if !reflect.DeepEqual(newLeaderRoot, root) {
+			r.Fatalf("got %v, want %v", newLeaderRoot, root)
+		}
+	})
+}
+
+func TestLeader_ACLUpgrade(t *testing.T) {
+	t.Parallel()
+	dir1, s1 := testServerWithConfig(t, func(c *Config) {
+		c.ACLsEnabled = true
+		c.ACLMasterToken = "root"
+	})
+	defer os.RemoveAll(dir1)
+	defer s1.Shutdown()
+	testrpc.WaitForTestAgent(t, s1.RPC, "dc1")
+	codec := rpcClient(t, s1)
+	defer codec.Close()
+
+	// create a legacy management ACL
+	mgmt := structs.ACLRequest{
+		Datacenter: "dc1",
+		Op:         structs.ACLSet,
+		ACL: structs.ACL{
+			Name: "Management token",
+			Type: structs.ACLTokenTypeManagement,
+		},
+		WriteRequest: structs.WriteRequest{Token: "root"},
+	}
+	var mgmt_id string
+	require.NoError(t, msgpackrpc.CallWithCodec(codec, "ACL.Apply", &mgmt, &mgmt_id))
+
+	// wait for it to be upgraded
+	retry.Run(t, func(t *retry.R) {
+		_, token, err := s1.fsm.State().ACLTokenGetBySecret(nil, mgmt_id)
+		require.NoError(t, err)
+		require.NotNil(t, token)
+		require.NotEqual(t, "", token.AccessorID)
+		require.Equal(t, structs.ACLTokenTypeManagement, token.Type)
+		require.Len(t, token.Policies, 1)
+		require.Equal(t, structs.ACLPolicyGlobalManagementID, token.Policies[0].ID)
+	})
+
+	// create a legacy management ACL
+	client := structs.ACLRequest{
+		Datacenter: "dc1",
+		Op:         structs.ACLSet,
+		ACL: structs.ACL{
+			Name:  "Management token",
+			Type:  structs.ACLTokenTypeClient,
+			Rules: `node "" { policy = "read"}`,
+		},
+		WriteRequest: structs.WriteRequest{Token: "root"},
+	}
+	var client_id string
+	require.NoError(t, msgpackrpc.CallWithCodec(codec, "ACL.Apply", &client, &client_id))
+
+	// wait for it to be upgraded
+	retry.Run(t, func(t *retry.R) {
+		_, token, err := s1.fsm.State().ACLTokenGetBySecret(nil, client_id)
+		require.NoError(t, err)
+		require.NotNil(t, token)
+		require.NotEqual(t, "", token.AccessorID)
+		require.Len(t, token.Policies, 0)
+		require.Equal(t, structs.ACLTokenTypeClient, token.Type)
+		require.Equal(t, client.ACL.Rules, token.Rules)
+	})
+}
+
+func TestLeader_ConfigEntryBootstrap(t *testing.T) {
+	t.Parallel()
+	global_entry_init := &structs.ProxyConfigEntry{
+		Kind: structs.ProxyDefaults,
+		Name: structs.ProxyConfigGlobal,
+		Config: map[string]interface{}{
+			"foo": "bar",
+			"bar": int64(1),
+		},
+	}
+
+	dir1, s1 := testServerWithConfig(t, func(c *Config) {
+		c.Build = "1.5.0"
+		c.ConfigEntryBootstrap = []structs.ConfigEntry{
+			global_entry_init,
+		}
+	})
+	defer os.RemoveAll(dir1)
+	defer s1.Shutdown()
+	testrpc.WaitForTestAgent(t, s1.RPC, "dc1")
+
+	retry.Run(t, func(t *retry.R) {
+		_, entry, err := s1.fsm.State().ConfigEntry(nil, structs.ProxyDefaults, structs.ProxyConfigGlobal)
+		require.NoError(t, err)
+		require.NotNil(t, entry)
+		global, ok := entry.(*structs.ProxyConfigEntry)
+		require.True(t, ok)
+		require.Equal(t, global_entry_init.Kind, global.Kind)
+		require.Equal(t, global_entry_init.Name, global.Name)
+		require.Equal(t, global_entry_init.Config, global.Config)
+	})
+}
+
+func TestLeader_ParseCARoot(t *testing.T) {
+	type test struct {
+		pem           string
+		expectedError bool
+	}
+	tests := []test{
+		{"", true},
+		{`-----BEGIN CERTIFICATE-----
+MIIDHDCCAsKgAwIBAgIQS+meruRVzrmVwEhXNrtk9jAKBggqhkjOPQQDAjCBuTEL
+MAkGA1UEBhMCVVMxCzAJBgNVBAgTAkNBMRYwFAYDVQQHEw1TYW4gRnJhbmNpc2Nv
+MRowGAYDVQQJExExMDEgU2Vjb25kIFN0cmVldDEOMAwGA1UEERMFOTQxMDUxFzAV
+BgNVBAoTDkhhc2hpQ29ycCBJbmMuMUAwPgYDVQQDEzdDb25zdWwgQWdlbnQgQ0Eg
+MTkzNzYxNzQwMjcxNzUxOTkyMzAyMzE1NDkxNjUzODYyMzAwNzE3MB4XDTE5MDQx
+MjA5MTg0NVoXDTIwMDQxMTA5MTg0NVowHDEaMBgGA1UEAxMRY2xpZW50LmRjMS5j
+b25zdWwwWTATBgcqhkjOPQIBBggqhkjOPQMBBwNCAAS2UroGUh5k7eR//iPsn9ne
+CMCVsERnjqQnK6eDWnM5kTXgXcPPe5pcAS9xs0g8BZ+oVsJSc7sH6RYvX+gw6bCl
+o4IBRjCCAUIwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGAQUFBwMCBggr
+BgEFBQcDATAMBgNVHRMBAf8EAjAAMGgGA1UdDgRhBF84NDphNDplZjoxYTpjODo1
+MzoxMDo1YTpjNTplYTpjZTphYTowZDo2ZjpjOTozODozZDphZjo0NTphZTo5OTo4
+YzpiYjoyNzpiYzpiMzpmYTpmMDozMToxNDo4ZTozNDBqBgNVHSMEYzBhgF8yYTox
+MjpjYTo0Mzo0NzowODpiZjoxYTo0Yjo4MTpkNDo2MzowNTo1ODowZToxYzo3Zjoy
+NTo0ZjozNDpmNDozYjpmYzo5YTpkNzo4Mjo2YjpkYzpmODo3YjphMTo5ZDAtBgNV
+HREEJjAkghFjbGllbnQuZGMxLmNvbnN1bIIJbG9jYWxob3N0hwR/AAABMAoGCCqG
+SM49BAMCA0gAMEUCIHcLS74KSQ7RA+edwOprmkPTh1nolwXz9/y9CJ5nMVqEAiEA
+h1IHCbxWsUT3AiARwj5/D/CUppy6BHIFkvcpOCQoVyo=
+-----END CERTIFICATE-----`, false},
+	}
+	for _, test := range tests {
+		root, err := parseCARoot(test.pem, "consul", "cluster")
+		if err == nil && test.expectedError {
+			require.Error(t, err)
+		}
+		if test.pem != "" {
+			rootCert, err := connect.ParseCert(test.pem)
+			require.NoError(t, err)
+
+			// just to make sure these two are not the same
+			require.NotEqual(t, rootCert.AuthorityKeyId, rootCert.SubjectKeyId)
+
+			require.Equal(t, connect.HexString(rootCert.SubjectKeyId), root.SigningKeyID)
+		}
+	}
+}
