@@ -18,7 +18,7 @@ func (op *Operator) RaftGetConfiguration(args *structs.DCSpecificRequest, reply
 	}
 
 	// This action requires operator read access.
-	rule, err := op.srv.resolveToken(args.Token)
+	rule, err := op.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
@@ -80,7 +80,7 @@ func (op *Operator) RaftRemovePeerByAddress(args *structs.RaftRemovePeerRequest,
 
 	// This is a super dangerous operation that requires operator write
 	// access.
-	rule, err := op.srv.resolveToken(args.Token)
+	rule, err := op.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
@@ -147,7 +147,7 @@ func (op *Operator) RaftRemovePeerByID(args *structs.RaftRemovePeerRequest, repl
 
 	// This is a super dangerous operation that requires operator write
 	// access.
-	rule, err := op.srv.resolveToken(args.Token)
+	rule, err := op.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
