@@ -32,7 +32,6 @@ func (p *PreparedQuery) Apply(args *structs.PreparedQueryRequest, reply *string)
 	if done, err := p.srv.forward("PreparedQuery.Apply", args, args, reply); done {
 		return err
 	}
-	defer metrics.MeasureSince([]string{"consul", "prepared-query", "apply"}, time.Now())
 	defer metrics.MeasureSince([]string{"prepared-query", "apply"}, time.Now())
 
 	// Validate the ID. We must create new IDs before applying to the Raft
@@ -61,7 +60,7 @@ func (p *PreparedQuery) Apply(args *structs.PreparedQueryRequest, reply *string)
 	*reply = args.Query.ID
 
 	// Get the ACL token for the request for the checks below.
-	rule, err := p.srv.resolveToken(args.Token)
+	rule, err := p.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
@@ -287,7 +286,6 @@ func (p *PreparedQuery) Explain(args *structs.PreparedQueryExecuteRequest,
 	if done, err := p.srv.forward("PreparedQuery.Explain", args, args, reply); done {
 		return err
 	}
-	defer metrics.MeasureSince([]string{"consul", "prepared-query", "explain"}, time.Now())
 	defer metrics.MeasureSince([]string{"prepared-query", "explain"}, time.Now())
 
 	// We have to do this ourselves since we are not doing a blocking RPC.
@@ -335,7 +333,6 @@ func (p *PreparedQuery) Execute(args *structs.PreparedQueryExecuteRequest,
 	if done, err := p.srv.forward("PreparedQuery.Execute", args, args, reply); done {
 		return err
 	}
-	defer metrics.MeasureSince([]string{"consul", "prepared-query", "execute"}, time.Now())
 	defer metrics.MeasureSince([]string{"prepared-query", "execute"}, time.Now())
 
 	// We have to do this ourselves since we are not doing a blocking RPC.
@@ -357,7 +354,7 @@ func (p *PreparedQuery) Execute(args *structs.PreparedQueryExecuteRequest,
 	}
 
 	// Execute the query for the local DC.
-	if err := p.execute(query, reply); err != nil {
+	if err := p.execute(query, reply, args.Connect); err != nil {
 		return err
 	}
 
@@ -453,7 +450,7 @@ func (p *PreparedQuery) Execute(args *structs.PreparedQueryExecuteRequest,
 	// by the query setup.
 	if len(reply.Nodes) == 0 {
 		wrapper := &queryServerWrapper{p.srv}
-		if err := queryFailover(wrapper, query, args.Limit, args.QueryOptions, reply); err != nil {
+		if err := queryFailover(wrapper, query, args, reply); err != nil {
 			return err
 		}
 	}
@@ -471,7 +468,6 @@ func (p *PreparedQuery) ExecuteRemote(args *structs.PreparedQueryExecuteRemoteRe
 	if done, err := p.srv.forward("PreparedQuery.ExecuteRemote", args, args, reply); done {
 		return err
 	}
-	defer metrics.MeasureSince([]string{"consul", "prepared-query", "execute_remote"}, time.Now())
 	defer metrics.MeasureSince([]string{"prepared-query", "execute_remote"}, time.Now())
 
 	// We have to do this ourselves since we are not doing a blocking RPC.
@@ -483,7 +479,7 @@ func (p *PreparedQuery) ExecuteRemote(args *structs.PreparedQueryExecuteRemoteRe
 	}
 
 	// Run the query locally to see what we can find.
-	if err := p.execute(&args.Query, reply); err != nil {
+	if err := p.execute(&args.Query, reply, args.Connect); err != nil {
 		return err
 	}
 
@@ -513,9 +509,18 @@ func (p *PreparedQuery) ExecuteRemote(args *structs.PreparedQueryExecuteRemoteRe
 // execute runs a prepared query in the local DC without any failover. We don't
 // apply any sorting options or ACL checks at this level - it should be done up above.
 func (p *PreparedQuery) execute(query *structs.PreparedQuery,
-	reply *structs.PreparedQueryExecuteResponse) error {
+	reply *structs.PreparedQueryExecuteResponse,
+	forceConnect bool) error {
 	state := p.srv.fsm.State()
-	_, nodes, err := state.CheckServiceNodes(nil, query.Service.Service)
+
+	// If we're requesting Connect-capable services, then switch the
+	// lookup to be the Connect function.
+	f := state.CheckServiceNodes
+	if query.Service.Connect || forceConnect {
+		f = state.CheckConnectServiceNodes
+	}
+
+	_, nodes, err := f(nil, query.Service.Service)
 	if err != nil {
 		return err
 	}
@@ -529,6 +534,11 @@ func (p *PreparedQuery) execute(query *structs.PreparedQuery,
 		nodes = nodeMetaFilter(query.Service.NodeMeta, nodes)
 	}
 
+	// Apply the service metadata filters, if any.
+	if len(query.Service.ServiceMeta) > 0 {
+		nodes = serviceMetaFilter(query.Service.ServiceMeta, nodes)
+	}
+
 	// Apply the tag filters, if any.
 	if len(query.Service.Tags) > 0 {
 		nodes = tagFilter(query.Service.Tags, nodes)
@@ -611,6 +621,16 @@ func nodeMetaFilter(filters map[string]string, nodes structs.CheckServiceNodes)
 	return filtered
 }
 
+func serviceMetaFilter(filters map[string]string, nodes structs.CheckServiceNodes) structs.CheckServiceNodes {
+	var filtered structs.CheckServiceNodes
+	for _, node := range nodes {
+		if structs.SatisfiesMetaFilters(node.Service.Meta, filters) {
+			filtered = append(filtered, node)
+		}
+	}
+	return filtered
+}
+
 // queryServer is a wrapper that makes it easier to test the failover logic.
 type queryServer interface {
 	GetLogger() *log.Logger
@@ -655,7 +675,7 @@ func (q *queryServerWrapper) ForwardDC(method, dc string, args interface{}, repl
 // queryFailover runs an algorithm to determine which DCs to try and then calls
 // them to try to locate alternative services.
 func queryFailover(q queryServer, query *structs.PreparedQuery,
-	limit int, options structs.QueryOptions,
+	args *structs.PreparedQueryExecuteRequest,
 	reply *structs.PreparedQueryExecuteResponse) error {
 
 	// Pull the list of other DCs. This is sorted by RTT in case the user
@@ -723,8 +743,9 @@ func queryFailover(q queryServer, query *structs.PreparedQuery,
 		remote := &structs.PreparedQueryExecuteRemoteRequest{
 			Datacenter:   dc,
 			Query:        *query,
-			Limit:        limit,
-			QueryOptions: options,
+			Limit:        args.Limit,
+			QueryOptions: args.QueryOptions,
+			Connect:      args.Connect,
 		}
 		if err := q.ForwardDC("PreparedQuery.ExecuteRemote", dc, remote, reply); err != nil {
 			q.GetLogger().Printf("[WARN] consul.prepared_query: Failed querying for service '%s' in datacenter '%s': %s", query.Service.Service, dc, err)
