@@ -3,26 +3,35 @@ package local_test
 import (
 	"errors"
 	"fmt"
+	"log"
+	"os"
 	"reflect"
 	"testing"
 	"time"
 
+	"github.com/hashicorp/consul/testrpc"
+
+	"github.com/hashicorp/go-memdb"
+
 	"github.com/hashicorp/consul/agent"
 	"github.com/hashicorp/consul/agent/config"
 	"github.com/hashicorp/consul/agent/local"
 	"github.com/hashicorp/consul/agent/structs"
 	"github.com/hashicorp/consul/agent/token"
 	"github.com/hashicorp/consul/api"
-	"github.com/hashicorp/consul/testutil/retry"
+	"github.com/hashicorp/consul/sdk/testutil/retry"
 	"github.com/hashicorp/consul/types"
 	"github.com/pascaldekloe/goe/verify"
+	"github.com/stretchr/testify/assert"
+	"github.com/stretchr/testify/require"
 )
 
 func TestAgentAntiEntropy_Services(t *testing.T) {
 	t.Parallel()
 	a := &agent.TestAgent{Name: t.Name()}
-	a.Start()
+	a.Start(t)
 	defer a.Shutdown()
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
 
 	// Register info
 	args := &structs.RegisterRequest{
@@ -38,6 +47,10 @@ func TestAgentAntiEntropy_Services(t *testing.T) {
 		Service: "mysql",
 		Tags:    []string{"master"},
 		Port:    5000,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv1, "")
 	args.Service = srv1
@@ -51,6 +64,10 @@ func TestAgentAntiEntropy_Services(t *testing.T) {
 		Service: "redis",
 		Tags:    []string{},
 		Port:    8000,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 0,
+		},
 	}
 	a.State.AddService(srv2, "")
 
@@ -68,6 +85,10 @@ func TestAgentAntiEntropy_Services(t *testing.T) {
 		Service: "web",
 		Tags:    []string{},
 		Port:    80,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv3, "")
 
@@ -77,6 +98,10 @@ func TestAgentAntiEntropy_Services(t *testing.T) {
 		Service: "lb",
 		Tags:    []string{},
 		Port:    443,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 0,
+		},
 	}
 	args.Service = srv4
 	if err := a.RPC("Catalog.Register", args, &out); err != nil {
@@ -90,6 +115,10 @@ func TestAgentAntiEntropy_Services(t *testing.T) {
 		Tags:    []string{},
 		Address: "127.0.0.10",
 		Port:    8000,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv5, "")
 
@@ -107,6 +136,10 @@ func TestAgentAntiEntropy_Services(t *testing.T) {
 		Service: "cache",
 		Tags:    []string{},
 		Port:    11211,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 0,
+		},
 	}
 	a.State.SetServiceState(&local.ServiceState{
 		Service: srv6,
@@ -224,11 +257,257 @@ func TestAgentAntiEntropy_Services(t *testing.T) {
 	}
 }
 
+func TestAgentAntiEntropy_Services_ConnectProxy(t *testing.T) {
+	t.Parallel()
+
+	assert := assert.New(t)
+	a := &agent.TestAgent{Name: t.Name()}
+	a.Start(t)
+	defer a.Shutdown()
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
+
+	// Register node info
+	var out struct{}
+	args := &structs.RegisterRequest{
+		Datacenter: "dc1",
+		Node:       a.Config.NodeName,
+		Address:    "127.0.0.1",
+	}
+
+	// Exists both same (noop)
+	srv1 := &structs.NodeService{
+		Kind:    structs.ServiceKindConnectProxy,
+		ID:      "mysql-proxy",
+		Service: "mysql-proxy",
+		Port:    5000,
+		Proxy:   structs.ConnectProxyConfig{DestinationServiceName: "db"},
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
+	}
+	a.State.AddService(srv1, "")
+	args.Service = srv1
+	assert.Nil(a.RPC("Catalog.Register", args, &out))
+
+	// Exists both, different (update)
+	srv2 := &structs.NodeService{
+		ID:      "redis-proxy",
+		Service: "redis-proxy",
+		Port:    8000,
+		Kind:    structs.ServiceKindConnectProxy,
+		Proxy:   structs.ConnectProxyConfig{DestinationServiceName: "redis"},
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 0,
+		},
+	}
+	a.State.AddService(srv2, "")
+
+	srv2_mod := new(structs.NodeService)
+	*srv2_mod = *srv2
+	srv2_mod.Port = 9000
+	args.Service = srv2_mod
+	assert.Nil(a.RPC("Catalog.Register", args, &out))
+
+	// Exists local (create)
+	srv3 := &structs.NodeService{
+		ID:      "web-proxy",
+		Service: "web-proxy",
+		Port:    80,
+		Kind:    structs.ServiceKindConnectProxy,
+		Proxy:   structs.ConnectProxyConfig{DestinationServiceName: "web"},
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
+	}
+	a.State.AddService(srv3, "")
+
+	// Exists remote (delete)
+	srv4 := &structs.NodeService{
+		ID:      "lb-proxy",
+		Service: "lb-proxy",
+		Port:    443,
+		Kind:    structs.ServiceKindConnectProxy,
+		Proxy:   structs.ConnectProxyConfig{DestinationServiceName: "lb"},
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 0,
+		},
+	}
+	args.Service = srv4
+	assert.Nil(a.RPC("Catalog.Register", args, &out))
+
+	// Exists local, in sync, remote missing (create)
+	srv5 := &structs.NodeService{
+		ID:      "cache-proxy",
+		Service: "cache-proxy",
+		Port:    11211,
+		Kind:    structs.ServiceKindConnectProxy,
+		Proxy:   structs.ConnectProxyConfig{DestinationServiceName: "cache-proxy"},
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
+	}
+	a.State.SetServiceState(&local.ServiceState{
+		Service: srv5,
+		InSync:  true,
+	})
+
+	assert.Nil(a.State.SyncFull())
+
+	var services structs.IndexedNodeServices
+	req := structs.NodeSpecificRequest{
+		Datacenter: "dc1",
+		Node:       a.Config.NodeName,
+	}
+	assert.Nil(a.RPC("Catalog.NodeServices", &req, &services))
+
+	// We should have 5 services (consul included)
+	assert.Len(services.NodeServices.Services, 5)
+
+	// All the services should match
+	for id, serv := range services.NodeServices.Services {
+		serv.CreateIndex, serv.ModifyIndex = 0, 0
+		switch id {
+		case "mysql-proxy":
+			assert.Equal(srv1, serv)
+		case "redis-proxy":
+			assert.Equal(srv2, serv)
+		case "web-proxy":
+			assert.Equal(srv3, serv)
+		case "cache-proxy":
+			assert.Equal(srv5, serv)
+		case structs.ConsulServiceID:
+			// ignore
+		default:
+			t.Fatalf("unexpected service: %v", id)
+		}
+	}
+
+	assert.Nil(servicesInSync(a.State, 4))
+
+	// Remove one of the services
+	a.State.RemoveService("cache-proxy")
+	assert.Nil(a.State.SyncFull())
+	assert.Nil(a.RPC("Catalog.NodeServices", &req, &services))
+
+	// We should have 4 services (consul included)
+	assert.Len(services.NodeServices.Services, 4)
+
+	// All the services should match
+	for id, serv := range services.NodeServices.Services {
+		serv.CreateIndex, serv.ModifyIndex = 0, 0
+		switch id {
+		case "mysql-proxy":
+			assert.Equal(srv1, serv)
+		case "redis-proxy":
+			assert.Equal(srv2, serv)
+		case "web-proxy":
+			assert.Equal(srv3, serv)
+		case structs.ConsulServiceID:
+			// ignore
+		default:
+			t.Fatalf("unexpected service: %v", id)
+		}
+	}
+
+	assert.Nil(servicesInSync(a.State, 3))
+}
+
+func TestAgent_ServiceWatchCh(t *testing.T) {
+	t.Parallel()
+	a := &agent.TestAgent{Name: t.Name()}
+	a.Start(t)
+	defer a.Shutdown()
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
+
+	require := require.New(t)
+
+	// register a local service
+	srv1 := &structs.NodeService{
+		ID:      "svc_id1",
+		Service: "svc1",
+		Tags:    []string{"tag1"},
+		Port:    6100,
+	}
+	require.NoError(a.State.AddService(srv1, ""))
+
+	verifyState := func(ss *local.ServiceState) {
+		require.NotNil(ss)
+		require.NotNil(ss.WatchCh)
+
+		// Sanity check WatchCh blocks
+		select {
+		case <-ss.WatchCh:
+			t.Fatal("should block until service changes")
+		default:
+		}
+	}
+
+	// Should be able to get a ServiceState
+	ss := a.State.ServiceState(srv1.ID)
+	verifyState(ss)
+
+	// Update service in another go routine
+	go func() {
+		srv2 := srv1
+		srv2.Port = 6200
+		require.NoError(a.State.AddService(srv2, ""))
+	}()
+
+	// We should observe WatchCh close
+	select {
+	case <-ss.WatchCh:
+		// OK!
+	case <-time.After(500 * time.Millisecond):
+		t.Fatal("timeout waiting for WatchCh to close")
+	}
+
+	// Should also fire for state being set explicitly
+	ss = a.State.ServiceState(srv1.ID)
+	verifyState(ss)
+
+	go func() {
+		a.State.SetServiceState(&local.ServiceState{
+			Service: ss.Service,
+			Token:   "foo",
+		})
+	}()
+
+	// We should observe WatchCh close
+	select {
+	case <-ss.WatchCh:
+		// OK!
+	case <-time.After(500 * time.Millisecond):
+		t.Fatal("timeout waiting for WatchCh to close")
+	}
+
+	// Should also fire for service being removed
+	ss = a.State.ServiceState(srv1.ID)
+	verifyState(ss)
+
+	go func() {
+		require.NoError(a.State.RemoveService(srv1.ID))
+	}()
+
+	// We should observe WatchCh close
+	select {
+	case <-ss.WatchCh:
+		// OK!
+	case <-time.After(500 * time.Millisecond):
+		t.Fatal("timeout waiting for WatchCh to close")
+	}
+}
+
 func TestAgentAntiEntropy_EnableTagOverride(t *testing.T) {
 	t.Parallel()
 	a := &agent.TestAgent{Name: t.Name()}
-	a.Start()
+	a.Start(t)
 	defer a.Shutdown()
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
 
 	args := &structs.RegisterRequest{
 		Datacenter: "dc1",
@@ -244,6 +523,10 @@ func TestAgentAntiEntropy_EnableTagOverride(t *testing.T) {
 		Tags:              []string{"tag1"},
 		Port:              6100,
 		EnableTagOverride: true,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv1, "")
 
@@ -254,6 +537,10 @@ func TestAgentAntiEntropy_EnableTagOverride(t *testing.T) {
 		Tags:              []string{"tag2"},
 		Port:              6200,
 		EnableTagOverride: false,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv2, "")
 
@@ -271,6 +558,10 @@ func TestAgentAntiEntropy_EnableTagOverride(t *testing.T) {
 		Tags:              []string{"tag1_mod"},
 		Port:              7100,
 		EnableTagOverride: true,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	if err := a.RPC("Catalog.Register", args, &out); err != nil {
 		t.Fatalf("err: %v", err)
@@ -282,6 +573,10 @@ func TestAgentAntiEntropy_EnableTagOverride(t *testing.T) {
 		Tags:              []string{"tag2_mod"},
 		Port:              7200,
 		EnableTagOverride: false,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 0,
+		},
 	}
 	if err := a.RPC("Catalog.Register", args, &out); err != nil {
 		t.Fatalf("err: %v", err)
@@ -315,6 +610,10 @@ func TestAgentAntiEntropy_EnableTagOverride(t *testing.T) {
 				Tags:              []string{"tag1_mod"},
 				Port:              6100,
 				EnableTagOverride: true,
+				Weights: &structs.Weights{
+					Passing: 1,
+					Warning: 1,
+				},
 			}
 			if !verify.Values(t, "", got, want) {
 				t.FailNow()
@@ -338,8 +637,9 @@ func TestAgentAntiEntropy_EnableTagOverride(t *testing.T) {
 
 func TestAgentAntiEntropy_Services_WithChecks(t *testing.T) {
 	t.Parallel()
-	a := agent.NewTestAgent(t.Name(), "")
+	a := agent.NewTestAgent(t, t.Name(), "")
 	defer a.Shutdown()
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
 
 	{
 		// Single check
@@ -472,8 +772,9 @@ func TestAgentAntiEntropy_Services_ACLDeny(t *testing.T) {
 		acl_master_token = "root"
 		acl_default_policy = "deny"
 		acl_enforce_version_8 = true`}
-	a.Start()
+	a.Start(t)
 	defer a.Shutdown()
+	testrpc.WaitForLeader(t, a.RPC, "dc1")
 
 	// Create the ACL
 	arg := structs.ACLRequest{
@@ -481,7 +782,7 @@ func TestAgentAntiEntropy_Services_ACLDeny(t *testing.T) {
 		Op:         structs.ACLSet,
 		ACL: structs.ACL{
 			Name:  "User token",
-			Type:  structs.ACLTypeClient,
+			Type:  structs.ACLTokenTypeClient,
 			Rules: testRegisterRules,
 		},
 		WriteRequest: structs.WriteRequest{
@@ -499,6 +800,10 @@ func TestAgentAntiEntropy_Services_ACLDeny(t *testing.T) {
 		Service: "mysql",
 		Tags:    []string{"master"},
 		Port:    5000,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv1, token)
 
@@ -508,6 +813,10 @@ func TestAgentAntiEntropy_Services_ACLDeny(t *testing.T) {
 		Service: "api",
 		Tags:    []string{"foo"},
 		Port:    5001,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 0,
+		},
 	}
 	a.State.AddService(srv2, token)
 
@@ -610,9 +919,10 @@ func TestAgentAntiEntropy_Services_ACLDeny(t *testing.T) {
 func TestAgentAntiEntropy_Checks(t *testing.T) {
 	t.Parallel()
 	a := &agent.TestAgent{Name: t.Name()}
-	a.Start()
+	a.Start(t)
 	defer a.Shutdown()
 
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
 	// Register info
 	args := &structs.RegisterRequest{
 		Datacenter: "dc1",
@@ -802,21 +1112,24 @@ func TestAgentAntiEntropy_Checks(t *testing.T) {
 
 func TestAgentAntiEntropy_Checks_ACLDeny(t *testing.T) {
 	t.Parallel()
+	dc := "dc1"
 	a := &agent.TestAgent{Name: t.Name(), HCL: `
-		acl_datacenter = "dc1"
+		acl_datacenter = "` + dc + `"
 		acl_master_token = "root"
 		acl_default_policy = "deny"
 		acl_enforce_version_8 = true`}
-	a.Start()
+	a.Start(t)
 	defer a.Shutdown()
 
+	testrpc.WaitForLeader(t, a.RPC, dc)
+
 	// Create the ACL
 	arg := structs.ACLRequest{
-		Datacenter: "dc1",
+		Datacenter: dc,
 		Op:         structs.ACLSet,
 		ACL: structs.ACL{
 			Name:  "User token",
-			Type:  structs.ACLTypeClient,
+			Type:  structs.ACLTokenTypeClient,
 			Rules: testRegisterRules,
 		},
 		WriteRequest: structs.WriteRequest{
@@ -834,6 +1147,10 @@ func TestAgentAntiEntropy_Checks_ACLDeny(t *testing.T) {
 		Service: "mysql",
 		Tags:    []string{"master"},
 		Port:    5000,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv1, "root")
 	srv2 := &structs.NodeService{
@@ -841,6 +1158,10 @@ func TestAgentAntiEntropy_Checks_ACLDeny(t *testing.T) {
 		Service: "api",
 		Tags:    []string{"foo"},
 		Port:    5001,
+		Weights: &structs.Weights{
+			Passing: 1,
+			Warning: 1,
+		},
 	}
 	a.State.AddService(srv2, "root")
 
@@ -851,7 +1172,7 @@ func TestAgentAntiEntropy_Checks_ACLDeny(t *testing.T) {
 	// Verify that we are in sync
 	{
 		req := structs.NodeSpecificRequest{
-			Datacenter: "dc1",
+			Datacenter: dc,
 			Node:       a.Config.NodeName,
 			QueryOptions: structs.QueryOptions{
 				Token: "root",
@@ -921,7 +1242,7 @@ func TestAgentAntiEntropy_Checks_ACLDeny(t *testing.T) {
 
 	// Verify that we are in sync
 	req := structs.NodeSpecificRequest{
-		Datacenter: "dc1",
+		Datacenter: dc,
 		Node:       a.Config.NodeName,
 		QueryOptions: structs.QueryOptions{
 			Token: "root",
@@ -967,7 +1288,7 @@ func TestAgentAntiEntropy_Checks_ACLDeny(t *testing.T) {
 	// Verify that we are in sync
 	{
 		req := structs.NodeSpecificRequest{
-			Datacenter: "dc1",
+			Datacenter: dc,
 			Node:       a.Config.NodeName,
 			QueryOptions: structs.QueryOptions{
 				Token: "root",
@@ -1011,11 +1332,12 @@ func TestAgentAntiEntropy_Checks_ACLDeny(t *testing.T) {
 
 func TestAgent_UpdateCheck_DiscardOutput(t *testing.T) {
 	t.Parallel()
-	a := agent.NewTestAgent(t.Name(), `
+	a := agent.NewTestAgent(t, t.Name(), `
 		discard_check_output = true
 		check_update_interval = "0s" # set to "0s" since otherwise output checks are deferred
 	`)
 	defer a.Shutdown()
+	testrpc.WaitForLeader(t, a.RPC, "dc1")
 
 	inSync := func(id string) bool {
 		s := a.State.CheckState(types.CheckID(id))
@@ -1064,8 +1386,9 @@ func TestAgentAntiEntropy_Check_DeferSync(t *testing.T) {
 	a := &agent.TestAgent{Name: t.Name(), HCL: `
 		check_update_interval = "500ms"
 	`}
-	a.Start()
+	a.Start(t)
 	defer a.Shutdown()
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
 
 	// Create a check
 	check := &structs.HealthCheck{
@@ -1254,8 +1577,9 @@ func TestAgentAntiEntropy_NodeInfo(t *testing.T) {
 		node_meta {
 			somekey = "somevalue"
 		}`}
-	a.Start()
+	a.Start(t)
 	defer a.Shutdown()
+	testrpc.WaitForLeader(t, a.RPC, "dc1")
 
 	// Register info
 	args := &structs.RegisterRequest{
@@ -1322,7 +1646,7 @@ func TestAgent_ServiceTokens(t *testing.T) {
 	t.Parallel()
 
 	tokens := new(token.Store)
-	tokens.UpdateUserToken("default")
+	tokens.UpdateUserToken("default", token.TokenSourceConfig)
 	cfg := config.DefaultRuntimeConfig(`bind_addr = "127.0.0.1" data_dir = "dummy"`)
 	l := local.NewState(agent.LocalConfig(cfg), nil, tokens)
 	l.TriggerSyncChanges = func() {}
@@ -1351,7 +1675,7 @@ func TestAgent_CheckTokens(t *testing.T) {
 	t.Parallel()
 
 	tokens := new(token.Store)
-	tokens.UpdateUserToken("default")
+	tokens.UpdateUserToken("default", token.TokenSourceConfig)
 	cfg := config.DefaultRuntimeConfig(`bind_addr = "127.0.0.1" data_dir = "dummy"`)
 	l := local.NewState(agent.LocalConfig(cfg), nil, tokens)
 	l.TriggerSyncChanges = func() {}
@@ -1409,18 +1733,18 @@ func TestAgent_CheckCriticalTime(t *testing.T) {
 	if c, ok := l.CriticalCheckStates()[checkID]; !ok {
 		t.Fatalf("should have a critical check")
 	} else if c.CriticalFor() > time.Millisecond {
-		t.Fatalf("bad: %#v", c)
+		t.Fatalf("bad: %#v, check was critical for %v", c, c.CriticalFor())
 	}
 
 	// Wait a while, then fail it again and make sure the time keeps track
-	// of the initial failure, and doesn't reset here.
+	// of the initial failure, and doesn't reset here. Since we are sleeping for
+	// 50ms the check should not be any less than that.
 	time.Sleep(50 * time.Millisecond)
 	l.UpdateCheck(chk.CheckID, api.HealthCritical, "")
 	if c, ok := l.CriticalCheckStates()[checkID]; !ok {
 		t.Fatalf("should have a critical check")
-	} else if c.CriticalFor() < 25*time.Millisecond ||
-		c.CriticalFor() > 75*time.Millisecond {
-		t.Fatalf("bad: %#v", c)
+	} else if c.CriticalFor() < 50*time.Millisecond {
+		t.Fatalf("bad: %#v, check was critical for %v", c, c.CriticalFor())
 	}
 
 	// Set it passing again.
@@ -1435,7 +1759,7 @@ func TestAgent_CheckCriticalTime(t *testing.T) {
 	if c, ok := l.CriticalCheckStates()[checkID]; !ok {
 		t.Fatalf("should have a critical check")
 	} else if c.CriticalFor() > time.Millisecond {
-		t.Fatalf("bad: %#v", c)
+		t.Fatalf("bad: %#v, check was critical for %v", c, c.CriticalFor())
 	}
 }
 
@@ -1460,9 +1784,60 @@ func TestAgent_AddCheckFailure(t *testing.T) {
 	}
 }
 
+func TestAgent_AliasCheck(t *testing.T) {
+	t.Parallel()
+
+	require := require.New(t)
+	cfg := config.DefaultRuntimeConfig(`bind_addr = "127.0.0.1" data_dir = "dummy"`)
+	l := local.NewState(agent.LocalConfig(cfg), nil, new(token.Store))
+	l.TriggerSyncChanges = func() {}
+
+	// Add checks
+	require.NoError(l.AddService(&structs.NodeService{Service: "s1"}, ""))
+	require.NoError(l.AddService(&structs.NodeService{Service: "s2"}, ""))
+	require.NoError(l.AddCheck(&structs.HealthCheck{CheckID: types.CheckID("c1"), ServiceID: "s1"}, ""))
+	require.NoError(l.AddCheck(&structs.HealthCheck{CheckID: types.CheckID("c2"), ServiceID: "s2"}, ""))
+
+	// Add an alias
+	notifyCh := make(chan struct{}, 1)
+	require.NoError(l.AddAliasCheck(types.CheckID("a1"), "s1", notifyCh))
+
+	// Update and verify we get notified
+	l.UpdateCheck(types.CheckID("c1"), api.HealthCritical, "")
+	select {
+	case <-notifyCh:
+	default:
+		t.Fatal("notify not received")
+	}
+
+	// Update again and verify we do not get notified
+	l.UpdateCheck(types.CheckID("c1"), api.HealthCritical, "")
+	select {
+	case <-notifyCh:
+		t.Fatal("notify received")
+	default:
+	}
+
+	// Update other check and verify we do not get notified
+	l.UpdateCheck(types.CheckID("c2"), api.HealthCritical, "")
+	select {
+	case <-notifyCh:
+		t.Fatal("notify received")
+	default:
+	}
+
+	// Update change and verify we get notified
+	l.UpdateCheck(types.CheckID("c1"), api.HealthPassing, "")
+	select {
+	case <-notifyCh:
+	default:
+		t.Fatal("notify not received")
+	}
+}
+
 func TestAgent_sendCoordinate(t *testing.T) {
 	t.Parallel()
-	a := agent.NewTestAgent(t.Name(), `
+	a := agent.NewTestAgent(t, t.Name(), `
 		sync_coordinate_interval_min = "1ms"
 		sync_coordinate_rate_target = 10.0
 		consul = {
@@ -1474,6 +1849,7 @@ func TestAgent_sendCoordinate(t *testing.T) {
 		}
 	`)
 	defer a.Shutdown()
+	testrpc.WaitForLeader(t, a.RPC, "dc1")
 
 	t.Logf("%d %d %s",
 		a.Config.ConsulCoordinateUpdateBatchSize,
@@ -1524,3 +1900,408 @@ func checksInSync(state *local.State, wantChecks int) error {
 	}
 	return nil
 }
+
+func TestState_Notify(t *testing.T) {
+	t.Parallel()
+
+	state := local.NewState(local.Config{},
+		log.New(os.Stderr, "", log.LstdFlags), &token.Store{})
+
+	// Stub state syncing
+	state.TriggerSyncChanges = func() {}
+
+	require := require.New(t)
+	assert := assert.New(t)
+
+	// Register a notifier
+	notifyCh := make(chan struct{}, 1)
+	state.Notify(notifyCh)
+	defer state.StopNotify(notifyCh)
+	assert.Empty(notifyCh)
+	drainCh(notifyCh)
+
+	// Add a service
+	err := state.AddService(&structs.NodeService{
+		Service: "web",
+	}, "fake-token-web")
+	require.NoError(err)
+
+	// Should have a notification
+	assert.NotEmpty(notifyCh)
+	drainCh(notifyCh)
+
+	// Re-Add same service
+	err = state.AddService(&structs.NodeService{
+		Service: "web",
+		Port:    4444,
+	}, "fake-token-web")
+	require.NoError(err)
+
+	// Should have a notification
+	assert.NotEmpty(notifyCh)
+	drainCh(notifyCh)
+
+	// Remove service
+	require.NoError(state.RemoveService("web"))
+
+	// Should have a notification
+	assert.NotEmpty(notifyCh)
+	drainCh(notifyCh)
+
+	// Stopping should... stop
+	state.StopNotify(notifyCh)
+
+	// Add a service
+	err = state.AddService(&structs.NodeService{
+		Service: "web",
+	}, "fake-token-web")
+	require.NoError(err)
+
+	// Should NOT have a notification
+	assert.Empty(notifyCh)
+	drainCh(notifyCh)
+}
+
+func TestStateProxyManagement(t *testing.T) {
+	t.Parallel()
+
+	state := local.NewState(local.Config{
+		ProxyBindMinPort: 20000,
+		ProxyBindMaxPort: 20001,
+	}, log.New(os.Stderr, "", log.LstdFlags), &token.Store{})
+
+	// Stub state syncing
+	state.TriggerSyncChanges = func() {}
+
+	p1 := structs.ConnectManagedProxy{
+		ExecMode:        structs.ProxyExecModeDaemon,
+		Command:         []string{"consul", "connect", "proxy"},
+		TargetServiceID: "web",
+	}
+
+	require := require.New(t)
+	assert := assert.New(t)
+
+	_, err := state.AddProxy(&p1, "fake-token", "")
+	require.Error(err, "should fail as the target service isn't registered")
+
+	// Sanity check done, lets add a couple of target services to the state
+	err = state.AddService(&structs.NodeService{
+		Service: "web",
+	}, "fake-token-web")
+	require.NoError(err)
+	err = state.AddService(&structs.NodeService{
+		Service: "cache",
+	}, "fake-token-cache")
+	require.NoError(err)
+	require.NoError(err)
+	err = state.AddService(&structs.NodeService{
+		Service: "db",
+	}, "fake-token-db")
+	require.NoError(err)
+
+	// Should work now
+	pstate, err := state.AddProxy(&p1, "fake-token", "")
+	require.NoError(err)
+
+	svc := pstate.Proxy.ProxyService
+	assert.Equal("web-proxy", svc.ID)
+	assert.Equal("web-proxy", svc.Service)
+	assert.Equal(structs.ServiceKindConnectProxy, svc.Kind)
+	assert.Equal("web", svc.Proxy.DestinationServiceName)
+	assert.Equal("", svc.Address, "should have empty address by default")
+	// Port is non-deterministic but could be either of 20000 or 20001
+	assert.Contains([]int{20000, 20001}, svc.Port)
+
+	{
+		// Re-registering same proxy again should not pick a random port but re-use
+		// the assigned one. It should also keep the same proxy token since we don't
+		// want to force restart for config change.
+		pstateDup, err := state.AddProxy(&p1, "fake-token", "")
+		require.NoError(err)
+		svcDup := pstateDup.Proxy.ProxyService
+
+		assert.Equal("web-proxy", svcDup.ID)
+		assert.Equal("web-proxy", svcDup.Service)
+		assert.Equal(structs.ServiceKindConnectProxy, svcDup.Kind)
+		assert.Equal("web", svcDup.Proxy.DestinationServiceName)
+		assert.Equal("", svcDup.Address, "should have empty address by default")
+		// Port must be same as before
+		assert.Equal(svc.Port, svcDup.Port)
+		// Same ProxyToken
+		assert.Equal(pstate.ProxyToken, pstateDup.ProxyToken)
+	}
+
+	// Let's register a notifier now
+	notifyCh := make(chan struct{}, 1)
+	state.NotifyProxy(notifyCh)
+	defer state.StopNotifyProxy(notifyCh)
+	assert.Empty(notifyCh)
+	drainCh(notifyCh)
+
+	// Second proxy should claim other port
+	p2 := p1
+	p2.TargetServiceID = "cache"
+	pstate2, err := state.AddProxy(&p2, "fake-token", "")
+	require.NoError(err)
+	svc2 := pstate2.Proxy.ProxyService
+	assert.Contains([]int{20000, 20001}, svc2.Port)
+	assert.NotEqual(svc.Port, svc2.Port)
+
+	// Should have a notification
+	assert.NotEmpty(notifyCh)
+	drainCh(notifyCh)
+
+	// Store this for later
+	p2token := state.Proxy(svc2.ID).ProxyToken
+
+	// Third proxy should fail as all ports are used
+	p3 := p1
+	p3.TargetServiceID = "db"
+	_, err = state.AddProxy(&p3, "fake-token", "")
+	require.Error(err)
+
+	// Should have a notification but we'll do nothing so that the next
+	// receive should block (we set cap == 1 above)
+
+	// But if we set a port explicitly it should be OK
+	p3.Config = map[string]interface{}{
+		"bind_port":    1234,
+		"bind_address": "0.0.0.0",
+	}
+	pstate3, err := state.AddProxy(&p3, "fake-token", "")
+	require.NoError(err)
+	svc3 := pstate3.Proxy.ProxyService
+	require.Equal("0.0.0.0", svc3.Address)
+	require.Equal(1234, svc3.Port)
+
+	// Should have a notification
+	assert.NotEmpty(notifyCh)
+	drainCh(notifyCh)
+
+	// Update config of an already registered proxy should work
+	p3updated := p3
+	p3updated.Config["foo"] = "bar"
+	// Setup multiple watchers who should all witness the change
+	gotP3 := state.Proxy(svc3.ID)
+	require.NotNil(gotP3)
+	var ws memdb.WatchSet
+	ws.Add(gotP3.WatchCh)
+	pstate3, err = state.AddProxy(&p3updated, "fake-token", "")
+	require.NoError(err)
+	svc3 = pstate3.Proxy.ProxyService
+	require.Equal("0.0.0.0", svc3.Address)
+	require.Equal(1234, svc3.Port)
+	gotProxy3 := state.Proxy(svc3.ID)
+	require.NotNil(gotProxy3)
+	require.Equal(p3updated.Config, gotProxy3.Proxy.Config)
+	assert.False(ws.Watch(time.After(500*time.Millisecond)),
+		"watch should have fired so ws.Watch should not timeout")
+
+	drainCh(notifyCh)
+
+	// Remove one of the auto-assigned proxies
+	_, err = state.RemoveProxy(svc2.ID)
+	require.NoError(err)
+
+	// Should have a notification
+	assert.NotEmpty(notifyCh)
+	drainCh(notifyCh)
+
+	// Should be able to create a new proxy for that service with the port (it
+	// should have been "freed").
+	p4 := p2
+	pstate4, err := state.AddProxy(&p4, "fake-token", "")
+	require.NoError(err)
+	svc4 := pstate4.Proxy.ProxyService
+	assert.Contains([]int{20000, 20001}, svc2.Port)
+	assert.Equal(svc4.Port, svc2.Port, "should get the same port back that we freed")
+
+	// Remove a proxy that doesn't exist should error
+	_, err = state.RemoveProxy("nope")
+	require.Error(err)
+
+	assert.Equal(&p4, state.Proxy(p4.ProxyService.ID).Proxy,
+		"should fetch the right proxy details")
+	assert.Nil(state.Proxy("nope"))
+
+	proxies := state.Proxies()
+	assert.Len(proxies, 3)
+	assert.Equal(&p1, proxies[svc.ID].Proxy)
+	assert.Equal(&p4, proxies[svc4.ID].Proxy)
+	assert.Equal(&p3, proxies[svc3.ID].Proxy)
+
+	tokens := make([]string, 4)
+	tokens[0] = state.Proxy(svc.ID).ProxyToken
+	// p2 not registered anymore but lets make sure p4 got a new token when it
+	// re-registered with same ID.
+	tokens[1] = p2token
+	tokens[2] = state.Proxy(svc2.ID).ProxyToken
+	tokens[3] = state.Proxy(svc3.ID).ProxyToken
+
+	// Quick check all are distinct
+	for i := 0; i < len(tokens)-1; i++ {
+		assert.Len(tokens[i], 36) // Sanity check for UUIDish thing.
+		for j := i + 1; j < len(tokens); j++ {
+			assert.NotEqual(tokens[i], tokens[j], "tokens for proxy %d and %d match",
+				i+1, j+1)
+		}
+	}
+}
+
+// Tests the logic for retaining tokens and ports through restore (i.e.
+// proxy-service already restored and token passed in externally)
+func TestStateProxyRestore(t *testing.T) {
+	t.Parallel()
+
+	state := local.NewState(local.Config{
+		// Wide random range to make it very unlikely to pass by chance
+		ProxyBindMinPort: 10000,
+		ProxyBindMaxPort: 20000,
+	}, log.New(os.Stderr, "", log.LstdFlags), &token.Store{})
+
+	// Stub state syncing
+	state.TriggerSyncChanges = func() {}
+
+	webSvc := structs.NodeService{
+		Service: "web",
+	}
+
+	p1 := structs.ConnectManagedProxy{
+		ExecMode:        structs.ProxyExecModeDaemon,
+		Command:         []string{"consul", "connect", "proxy"},
+		TargetServiceID: "web",
+	}
+
+	p2 := p1
+
+	require := require.New(t)
+	assert := assert.New(t)
+
+	// Add a target service
+	require.NoError(state.AddService(&webSvc, "fake-token-web"))
+
+	// Add the proxy for first time to get the proper service definition to
+	// register
+	pstate, err := state.AddProxy(&p1, "fake-token", "")
+	require.NoError(err)
+
+	// Now start again with a brand new state
+	state2 := local.NewState(local.Config{
+		// Wide random range to make it very unlikely to pass by chance
+		ProxyBindMinPort: 10000,
+		ProxyBindMaxPort: 20000,
+	}, log.New(os.Stderr, "", log.LstdFlags), &token.Store{})
+
+	// Stub state syncing
+	state2.TriggerSyncChanges = func() {}
+
+	// Register the target service
+	require.NoError(state2.AddService(&webSvc, "fake-token-web"))
+
+	// "Restore" the proxy service
+	require.NoError(state.AddService(p1.ProxyService, "fake-token-web"))
+
+	// Now we can AddProxy with the "restored" token
+	pstate2, err := state.AddProxy(&p2, "fake-token", pstate.ProxyToken)
+	require.NoError(err)
+
+	// Check it still has the same port and token as before
+	assert.Equal(pstate.ProxyToken, pstate2.ProxyToken)
+	assert.Equal(p1.ProxyService.Port, p2.ProxyService.Port)
+}
+
+// Test that alias check is updated after AddCheck, UpdateCheck, and RemoveCheck for the same service id
+func TestAliasNotifications_local(t *testing.T) {
+	t.Parallel()
+
+	a := agent.NewTestAgent(t, t.Name(), "")
+	defer a.Shutdown()
+
+	testrpc.WaitForTestAgent(t, a.RPC, "dc1")
+
+	// Register service with a failing TCP check
+	svcID := "socat"
+	srv := &structs.NodeService{
+		ID:      svcID,
+		Service: "echo",
+		Tags:    []string{},
+		Address: "127.0.0.10",
+		Port:    8080,
+	}
+	a.State.AddService(srv, "")
+
+	scID := "socat-sidecar-proxy"
+	sc := &structs.NodeService{
+		ID:      scID,
+		Service: scID,
+		Tags:    []string{},
+		Address: "127.0.0.10",
+		Port:    9090,
+	}
+	a.State.AddService(sc, "")
+
+	tcpID := types.CheckID("service:socat-tcp")
+	chk0 := &structs.HealthCheck{
+		Node:      "",
+		CheckID:   tcpID,
+		Name:      "tcp check",
+		Status:    api.HealthPassing,
+		ServiceID: svcID,
+	}
+	a.State.AddCheck(chk0, "")
+
+	// Register an alias for the service
+	proxyID := types.CheckID("service:socat-sidecar-proxy:2")
+	chk1 := &structs.HealthCheck{
+		Node:      "",
+		CheckID:   proxyID,
+		Name:      "Connect Sidecar Aliasing socat",
+		Status:    api.HealthPassing,
+		ServiceID: scID,
+	}
+	chkt := &structs.CheckType{
+		AliasService: svcID,
+	}
+	require.NoError(t, a.AddCheck(chk1, chkt, true, "", agent.ConfigSourceLocal))
+
+	// Add a failing check to the same service ID, alias should also fail
+	maintID := types.CheckID("service:socat-maintenance")
+	chk2 := &structs.HealthCheck{
+		Node:      "",
+		CheckID:   maintID,
+		Name:      "socat:Service Maintenance Mode",
+		Status:    api.HealthCritical,
+		ServiceID: svcID,
+	}
+	a.State.AddCheck(chk2, "")
+
+	retry.Run(t, func(r *retry.R) {
+		require.Equal(r, api.HealthCritical, a.State.Check(proxyID).Status)
+	})
+
+	// Remove the failing check, alias should pass
+	a.State.RemoveCheck(maintID)
+
+	retry.Run(t, func(r *retry.R) {
+		require.Equal(r, api.HealthPassing, a.State.Check(proxyID).Status)
+	})
+
+	// Update TCP check to failing, alias should fail
+	a.State.UpdateCheck(tcpID, api.HealthCritical, "")
+
+	retry.Run(t, func(r *retry.R) {
+		require.Equal(r, api.HealthCritical, a.State.Check(proxyID).Status)
+	})
+}
+
+// drainCh drains a channel by reading messages until it would block.
+func drainCh(ch chan struct{}) {
+	for {
+		select {
+		case <-ch:
+		default:
+			return
+		}
+	}
+}
