@@ -10,20 +10,24 @@ import (
 	"net/http/httptest"
 	"os"
 	"path/filepath"
-	"runtime"
 	"strconv"
 	"strings"
+	"testing"
 	"time"
 
 	metrics "github.com/armon/go-metrics"
+	uuid "github.com/hashicorp/go-uuid"
+
 	"github.com/hashicorp/consul/agent/config"
+	"github.com/hashicorp/consul/agent/connect"
 	"github.com/hashicorp/consul/agent/consul"
 	"github.com/hashicorp/consul/agent/structs"
 	"github.com/hashicorp/consul/api"
-	"github.com/hashicorp/consul/lib/freeport"
 	"github.com/hashicorp/consul/logger"
-	"github.com/hashicorp/consul/testutil/retry"
-	uuid "github.com/hashicorp/go-uuid"
+	"github.com/hashicorp/consul/sdk/freeport"
+	"github.com/hashicorp/consul/sdk/testutil/retry"
+
+	"github.com/stretchr/testify/require"
 )
 
 func init() {
@@ -43,6 +47,12 @@ type TestAgent struct {
 
 	HCL string
 
+	// ExpectConfigError can be set to prevent the agent retrying Start on errors
+	// and eventually blowing up with runtime.Goexit. This enables tests to assert
+	// that some specific bit of config actually does prevent startup entirely in
+	// a reasonable way without reproducing a lot of the boilerplate here.
+	ExpectConfigError bool
+
 	// Config is the agent configuration. If Config is nil then
 	// TestConfig() is used. If Config.DataDir is set then it is
 	// the callers responsibility to clean up the data directory.
@@ -83,25 +93,28 @@ type TestAgent struct {
 }
 
 // NewTestAgent returns a started agent with the given name and
-// configuration. It panics if the agent could not be started. The
+// configuration. It fails the test if the Agent could not be started. The
 // caller should call Shutdown() to stop the agent and remove temporary
 // directories.
-func NewTestAgent(name string, hcl string) *TestAgent {
+func NewTestAgent(t *testing.T, name string, hcl string) *TestAgent {
 	a := &TestAgent{Name: name, HCL: hcl}
-	a.Start()
+	a.Start(t)
 	return a
 }
 
-type panicFailer struct{}
-
-func (f *panicFailer) Log(args ...interface{}) { fmt.Println(args...) }
-func (f *panicFailer) FailNow()                { panic("failed") }
-
-// Start starts a test agent. It panics if the agent could not be started.
-func (a *TestAgent) Start() *TestAgent {
-	if a.Agent != nil {
-		panic("TestAgent already started")
+func NewUnstartedAgent(t *testing.T, name string, hcl string) (*Agent, error) {
+	c := TestConfig(config.Source{Name: name, Format: "hcl", Data: hcl})
+	a, err := New(c, nil)
+	if err != nil {
+		return nil, err
 	}
+	return a, nil
+}
+
+// Start starts a test agent. It fails the test if the agent could not be started.
+func (a *TestAgent) Start(t *testing.T) *TestAgent {
+	require := require.New(t)
+	require.Nil(a.Agent, "TestAgent already started")
 	var hclDataDir string
 	if a.DataDir == "" {
 		name := "agent"
@@ -110,44 +123,41 @@ func (a *TestAgent) Start() *TestAgent {
 		}
 		name = strings.Replace(name, "/", "_", -1)
 		d, err := ioutil.TempDir(TempDir, name)
-		if err != nil {
-			panic(fmt.Sprintf("Error creating data dir %s: %s", filepath.Join(TempDir, name), err))
-		}
+		require.NoError(err, fmt.Sprintf("Error creating data dir %s: %s", filepath.Join(TempDir, name), err))
 		hclDataDir = `data_dir = "` + d + `"`
 	}
-	id := NodeID()
 
+	var id string
 	for i := 10; i >= 0; i-- {
 		a.Config = TestConfig(
+			randomPortsSource(a.UseTLS),
 			config.Source{Name: a.Name, Format: "hcl", Data: a.HCL},
 			config.Source{Name: a.Name + ".data_dir", Format: "hcl", Data: hclDataDir},
-			randomPortsSource(a.UseTLS),
 		)
+		id = string(a.Config.NodeID)
 
 		// write the keyring
 		if a.Key != "" {
 			writeKey := func(key, filename string) {
 				path := filepath.Join(a.Config.DataDir, filename)
-				if err := initKeyring(path, key); err != nil {
-					panic(fmt.Sprintf("Error creating keyring %s: %s", path, err))
-				}
+				err := initKeyring(path, key)
+				require.NoError(err, fmt.Sprintf("Error creating keyring %s: %s", path, err))
 			}
 			writeKey(a.Key, SerfLANKeyring)
 			writeKey(a.Key, SerfWANKeyring)
 		}
 
-		agent, err := New(a.Config)
-		if err != nil {
-			panic(fmt.Sprintf("Error creating agent: %s", err))
-		}
-
 		logOutput := a.LogOutput
 		if logOutput == nil {
 			logOutput = os.Stderr
 		}
+		agentLogger := log.New(logOutput, a.Name+" - ", log.LstdFlags|log.Lmicroseconds)
+
+		agent, err := New(a.Config, agentLogger)
+		require.NoError(err, fmt.Sprintf("Error creating agent: %s", err))
+
 		agent.LogOutput = logOutput
 		agent.LogWriter = a.LogWriter
-		agent.logger = log.New(logOutput, a.Name+" - ", log.LstdFlags|log.Lmicroseconds)
 		agent.MemSink = metrics.NewInmemSink(1*time.Second, time.Minute)
 
 		// we need the err var in the next exit condition
@@ -155,8 +165,12 @@ func (a *TestAgent) Start() *TestAgent {
 			a.Agent = agent
 			break
 		} else if i == 0 {
-			fmt.Println(id, a.Name, "Error starting agent:", err)
-			runtime.Goexit()
+			require.Fail("%s %s Error starting agent: %s", id, a.Name, err)
+		} else if a.ExpectConfigError {
+			// Panic the error since this can be caught if needed. Pretty gross way to
+			// detect errors but enough for now and this is a tiny edge case that I'd
+			// otherwise not have a way to test at all...
+			panic(err)
 		} else {
 			agent.ShutdownAgent()
 			agent.ShutdownEndpoints()
@@ -170,8 +184,7 @@ func (a *TestAgent) Start() *TestAgent {
 		// the data dir, such as in the Raft configuration.
 		if a.DataDir != "" {
 			if err := os.RemoveAll(a.DataDir); err != nil {
-				fmt.Println(id, a.Name, "Error resetting data dir:", err)
-				runtime.Goexit()
+				require.Fail("%s %s Error resetting data dir: %s", id, a.Name, err)
 			}
 		}
 	}
@@ -180,7 +193,7 @@ func (a *TestAgent) Start() *TestAgent {
 	a.Agent.StartSync()
 
 	var out structs.IndexedNodes
-	retry.Run(&panicFailer{}, func(r *retry.R) {
+	retry.Run(t, func(r *retry.R) {
 		if len(a.httpServers) == 0 {
 			r.Fatal(a.Name, "waiting for server")
 		}
@@ -268,7 +281,8 @@ func (a *TestAgent) Client() *api.Client {
 // DNSDisableCompression disables compression for all started DNS servers.
 func (a *TestAgent) DNSDisableCompression(b bool) {
 	for _, srv := range a.dnsServers {
-		srv.disableCompression.Store(b)
+		cfg := srv.config.Load().(*dnsConfig)
+		cfg.DisableCompression = b
 	}
 }
 
@@ -334,6 +348,12 @@ func TestConfig(sources ...config.Source) *config.RuntimeConfig {
 			server = true
 			node_id = "` + nodeID + `"
 			node_name = "Node ` + nodeID + `"
+			connect {
+				enabled = true
+				ca_config {
+					cluster_id = "` + connect.TestClusterID + `"
+				}
+			}
 			performance {
 				raft_multiplier = 1
 			}
@@ -357,6 +377,14 @@ func TestConfig(sources ...config.Source) *config.RuntimeConfig {
 		fmt.Println("WARNING:", w)
 	}
 
+	// Disable connect proxy execution since it causes all kinds of problems with
+	// self-executing tests etc.
+	cfg.ConnectTestDisableManagedProxies = true
+	// Effectively disables the delay after root rotation before requesting CSRs
+	// to make test deterministic. 0 results in default jitter being applied but a
+	// tiny delay is effectively thre same.
+	cfg.ConnectTestCALeafRootChangeSpread = 1 * time.Nanosecond
+
 	return &cfg
 }
 
@@ -372,3 +400,18 @@ func TestACLConfig() string {
 		acl_enforce_version_8 = true
 	`
 }
+
+func TestACLConfigNew() string {
+	return `
+		primary_datacenter = "dc1"
+		acl {
+			enabled = true
+			default_policy = "deny"
+			tokens {
+				master = "root"
+				agent = "root"
+				agent_master = "towel"
+			}
+		}
+	`
+}
