@@ -17,7 +17,7 @@ func TestLeaveCommand_noTabs(t *testing.T) {
 
 func TestLeaveCommand(t *testing.T) {
 	t.Parallel()
-	a := agent.NewTestAgent(t.Name(), ``)
+	a := agent.NewTestAgent(t, t.Name(), ``)
 	defer a.Shutdown()
 
 	ui := cli.NewMockUi()
@@ -36,7 +36,7 @@ func TestLeaveCommand(t *testing.T) {
 
 func TestLeaveCommand_FailOnNonFlagArgs(t *testing.T) {
 	t.Parallel()
-	a := agent.NewTestAgent(t.Name(), ``)
+	a := agent.NewTestAgent(t, t.Name(), ``)
 	defer a.Shutdown()
 
 	ui := cli.NewMockUi()
