@@ -1,11 +1,12 @@
 import Mixin from '@ember/object/mixin';
+import { inject as service } from '@ember/service';
 import { get } from '@ember/object';
 import { assert } from '@ember/debug';
 export default Mixin.create({
+  dom: service('dom'),
   resize: function(e) {
     assert('with-resizing.resize needs to be overridden', false);
   },
-  win: window,
   init: function() {
     this._super(...arguments);
     this.handler = e => {
@@ -17,14 +18,18 @@ export default Mixin.create({
   },
   didInsertElement: function() {
     this._super(...arguments);
-    get(this, 'win').addEventListener('resize', this.handler, false);
+    get(this, 'dom')
+      .viewport()
+      .addEventListener('resize', this.handler, false);
     this.didAppear();
   },
   didAppear: function() {
-    this.handler({ target: get(this, 'win') });
+    this.handler({ target: get(this, 'dom').viewport() });
   },
   willDestroyElement: function() {
-    get(this, 'win').removeEventListener('resize', this.handler, false);
+    get(this, 'dom')
+      .viewport()
+      .removeEventListener('resize', this.handler, false);
     this._super(...arguments);
   },
 });
