@@ -5,6 +5,7 @@ import { get } from '@ember/object';
 
 export default Route.extend({
   repo: service('repository/service'),
+  settings: service('settings'),
   queryParams: {
     s: {
       as: 'filter',
@@ -13,19 +14,15 @@ export default Route.extend({
   },
   model: function(params) {
     const repo = get(this, 'repo');
+    const settings = get(this, 'settings');
+    const dc = this.modelFor('dc').dc.Name;
     return hash({
-      item: repo.findBySlug(params.name, this.modelFor('dc').dc.Name),
-    }).then(function(model) {
-      return {
-        ...model,
-        ...{
-          items: model.item.Nodes,
-        },
-      };
+      item: repo.findBySlug(params.name, dc),
+      urls: settings.findBySlug('urls'),
+      dc: dc,
     });
   },
   setupController: function(controller, model) {
-    this._super(...arguments);
     controller.setProperties(model);
   },
 });
