@@ -5,8 +5,12 @@ export default function(type) {
       requests = ['/v1/catalog/datacenters'];
       break;
     case 'service':
+    case 'instance':
       requests = ['/v1/internal/ui/services', '/v1/health/service/'];
       break;
+    case 'proxy':
+      requests = ['/v1/catalog/connect'];
+      break;
     case 'node':
       requests = ['/v1/internal/ui/nodes', '/v1/internal/ui/node/'];
       break;
@@ -22,6 +26,9 @@ export default function(type) {
     case 'policy':
       requests = ['/v1/acl/policies', '/v1/acl/policy/'];
       break;
+    case 'role':
+      requests = ['/v1/acl/roles', '/v1/acl/role/'];
+      break;
     case 'token':
       requests = ['/v1/acl/tokens', '/v1/acl/token/'];
       break;
