@@ -17,6 +17,10 @@ export default function(
         '[data-test-tabular-row]',
         deletable({
           id: attribute('data-test-token', '[data-test-token]'),
+          description: text('[data-test-description]'),
+          policy: text('[data-test-policy].policy', { multiple: true }),
+          role: text('[data-test-policy].role', { multiple: true }),
+          serviceIdentity: text('[data-test-policy].policy-service-identity', { multiple: true }),
           token: clickable('a'),
           actions: clickable('label'),
           use: clickable('[data-test-use]'),
