@@ -2,7 +2,7 @@ import { moduleFor, test } from 'ember-qunit';
 
 moduleFor('controller:dc/acls/policies/create', 'Unit | Controller | dc/acls/policies/create', {
   // Specify the other units that are required for this test.
-  needs: ['service:dom', 'service:form'],
+  needs: ['service:form', 'service:dom', 'service:repository/role', 'service:repository/policy'],
 });
 
 // Replace this with your real tests.
