@@ -2,7 +2,13 @@ import { moduleFor, test } from 'ember-qunit';
 
 moduleFor('controller:dc/kv/create', 'Unit | Controller | dc/kv/create', {
   // Specify the other units that are required for this test.
-  needs: ['service:btoa'],
+  needs: [
+    'service:btoa',
+    'service:dom',
+    'service:form',
+    'service:repository/role',
+    'service:repository/policy',
+  ],
 });
 
 // Replace this with your real tests.
